/* SelectMenu.java
 *
 * created: Thu Jan 14 1999
 *
 * This file is part of Artemis
 *
 * Copyright (C) 1998,1999  Genome Research Limited
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * $Header: /nfs/disk222/yeastpub/Repository/powmap/diana/components/SelectMenu.java,v 1.21 1999/09/09 11:24:40 kmr Exp $
 **/

package diana.components;

import diana.*;
import diana.sequence.*;

import uk.ac.sanger.pathogens.OutOfRangeException;
import uk.ac.sanger.pathogens.embl.Key;
import uk.ac.sanger.pathogens.embl.KeyVector;
import uk.ac.sanger.pathogens.embl.Qualifier;
import uk.ac.sanger.pathogens.embl.InvalidKeyException;
import uk.ac.sanger.pathogens.embl.LocationLexer;
import uk.ac.sanger.pathogens.embl.LocationLexer.*;

import java.awt.*;
import java.awt.event.*;

/**
 *  This menu has contains items such a "Select all", "Select none" and
 *  "Select by key".
 *
 *  @author Kim Rutherford
 *  @version $Id: SelectMenu.java,v 1.21 1999/09/09 11:24:40 kmr Exp $
 **/

public class SelectMenu extends SelectionMenu {
  /**
   *  Create a new SelectMenu object and all it's menu items.
   *  @param frame The Frame that owns this Menu.
   *  @param selection The Selection that the commands in the menu will
   *    operate on.
   *  @param goto_event_source The object the we will call makeBaseVisible ()
   *    on.
   *  @param entry_group The EntryGroup object where features and exons are
   *    selected from.
   **/
  public SelectMenu (final Frame frame,
                     final Selection selection,
                     final GotoEventSource goto_event_source,
                     final EntryGroup entry_group) {
    super (frame, "Select", selection);

    this.entry_group = entry_group;
    this.goto_event_source = goto_event_source;

    all_item = new MenuItem ("Select All",
                             new MenuShortcut (KeyEvent.VK_A));
    all_item.addActionListener (new ActionListener () {
      public void actionPerformed (ActionEvent event) {
        selectAll ();
      }
    });

    add (all_item);

    none_item = new MenuItem ("Select None",
                              new MenuShortcut (KeyEvent.VK_N));
    none_item.addActionListener (new ActionListener () {
      public void actionPerformed (ActionEvent event) {
        getSelection ().clear ();
      }
    });

    add (none_item);

    select_by_key_item = new MenuItem ("Select By Key");
    select_by_key_item.addActionListener (new ActionListener () {
      public void actionPerformed (ActionEvent event) {
        selectByKeyPopup ();
      }
    });

    add (select_by_key_item);

    select_cds_item = new MenuItem ("Select CDS Features",
                                    new MenuShortcut (KeyEvent.VK_F));
    select_cds_item.addActionListener (new ActionListener () {
      public void actionPerformed (ActionEvent event) {
        // select all CDS features that do not have the /pseudo qualifier
        final FeatureKeyQualifierPredicate predicate =
          new FeatureKeyQualifierPredicate (Key.CDS, "pseudo", false);

        getSelection ().clear ();

        selectFeaturesByPredicate (predicate);
      }
    });

    add (select_cds_item);

    select_same_type_item = new MenuItem ("Select Same Key");
    select_same_type_item.addActionListener (new ActionListener () {
      public void actionPerformed (ActionEvent event) {
        selectSameKey ();
      }
    });

    add (select_same_type_item);

    select_orf_item = new MenuItem ("Select Open Reading Frame");
    select_orf_item.addActionListener (new ActionListener () {
      public void actionPerformed (ActionEvent event) {
        selectOrf ();
      }
    });

    add (select_orf_item);

    select_range_item = new MenuItem ("Select Range ...");
    select_range_item.addActionListener (new ActionListener () {
      public void actionPerformed (ActionEvent event) {
        selectRange ();
      }
    });

    add (select_range_item);

    addSeparator ();

    selection_toggle_item = new MenuItem ("Toggle Selection",
                                          new MenuShortcut (KeyEvent.VK_T));
    selection_toggle_item.addActionListener (new ActionListener () {
      public void actionPerformed (ActionEvent event) {
        toggleSelection ();
      }
    });

    add (selection_toggle_item);
  }

  /**
   *  Select all the features in the active entries.  If there are more than
   *  1000 features the user is asked for confirmation.
   **/
  private void selectAll () {
    if (entry_group.getAllFeaturesCount () > 1000) {
      final YesNoDialog dialog =
        new YesNoDialog (getParentFrame (),
                         "Are you sure you want to select " +
                          entry_group.getAllFeaturesCount () + " features?");

      if (!dialog.getResult ()) {
        return;
      }
    }

    final Selection selection = getSelection ();
    final FeatureVector all_features = entry_group.getAllFeatures ();

    selection.set (all_features);
  }


  /**
   *  Invert the selection - after calling this method the selection will
   *  contain only those features that were not in the selection before the
   *  call.
   **/
  private void toggleSelection () {
    final FeatureVector selected_features = getSelection ().getAllFeatures ();

    final FeatureVector all_features = entry_group.getAllFeatures ();

    final FeatureVector new_selection = new FeatureVector ();

    for (int i = 0 ; i < all_features.size () ; ++i) {
      final Feature this_feature = all_features.elementAt (i);

      if (!selected_features.contains (this_feature)) {
        new_selection.add (this_feature);
      }
    }

    getSelection ().set (new_selection);
  }


  /**
   *  Popup a TextRequester component and then select all the features that
   *  have the same key as the user types.
   **/
  private void selectByKeyPopup () {
    final KeyChooser key_chooser;

    try {
      key_chooser = new KeyChooser (new Key ("misc_feature"));
    } catch (InvalidKeyException e) {
      key_chooser = new KeyChooser ();
    }

    key_chooser.getKeyChoice ().addItemListener (new ItemListener () {
      public void itemStateChanged (ItemEvent _) {
        getSelection ().clear ();
        selectByKey (key_chooser.getKeyChoice ().getSelectedItem ());
        key_chooser.setVisible (false);
        key_chooser.dispose ();
      }
    });

    key_chooser.getOKButton ().addActionListener (new ActionListener () {
      public void actionPerformed (ActionEvent _) {
        getSelection ().clear ();
        selectByKey (key_chooser.getKeyChoice ().getSelectedItem ());
        key_chooser.setVisible (false);
        key_chooser.dispose ();
      }
    });

    key_chooser.setVisible (true);
  }

  /**
   *  Select all the features that have the given key.
   **/
  private void selectByKey (final Key key) {
    final FeatureKeyPredicate predicate = new FeatureKeyPredicate (key);

    selectFeaturesByPredicate (predicate);
  }

  /**
   *  Select all the features that have the same key as the currently selected
   *  features.
   **/
  private void selectSameKey () {
    if (! checkForSelectionFeatures ()) {
      return;
    }

    final KeyVector seen_keys = new KeyVector ();

    final FeatureVector selected_features = getSelection ().getAllFeatures ();

    for (int i = 0 ; i < selected_features.size () ; ++i) {
      final Feature current_feature = selected_features.elementAt (i);

      if (!seen_keys.contains (current_feature.getKey ())) {
        seen_keys.add (current_feature.getKey ());
      }
    }

    getSelection ().clear ();

    for (int i = 0 ; i < seen_keys.size () ; ++i) {
      selectByKey (seen_keys.elementAt (i));
    }
  }

  /**
   *  Select all the features that match the given predicate.
   **/
  private void selectFeaturesByPredicate (final FeaturePredicate predicate) {
    final FeatureVector new_selection_features = new FeatureVector ();

    final FeatureEnumeration feature_enum = entry_group.features ();

    while (feature_enum.hasMoreFeatures ()) {
      final Feature current_feature = feature_enum.nextFeature ();

//      System.out.println ("adding a feature: " + new_feature.getKey ());

      if (predicate.testPredicate (current_feature)) {
        new_selection_features.add (current_feature);
      }
    }

    getSelection ().add (new_selection_features);
  }

  /**
   *  If a range of bases is selected, then select the ORF around those
   *  bases.
   **/
  private void selectOrf () {
    if (!checkForSelectionRange ()) {
      return;
    }

    final MarkerRange range = getSelection ().getMarkerRange ();

    final Strand strand = entry_group.getBases ().getForwardStrand ();

    final MarkerRange start_orf_range =
      strand.getORFAroundMarker (range.getStart ());

    final Marker end_marker;

    // get the marker of the first base of the last codon in the range (we
    // want to keep in the same frame)
    try {
      final int start_end_diff = range.getCount () - 1;

      final int mod_length = start_end_diff - start_end_diff % 3;

      end_marker = range.getStart ().moveBy (mod_length);
    } catch (OutOfRangeException e) {
      end_marker = range.getStart ();
    }

    final MarkerRange end_orf_range = strand.getORFAroundMarker (end_marker);

    MarkerRange new_range = range;

    if (start_orf_range != null) {
      new_range = new_range.extendRange (start_orf_range);
    }

    if (end_orf_range != null) {
      new_range = new_range.extendRange (end_orf_range);
    }

    if (start_orf_range == null && end_orf_range == null &&
        range.getCount () <= 6) {   // 6 == two codons
      new MessageDialog (getParentFrame (),
                         "there is no open reading frame at the selected " +
                         "bases");
      return;
    }

    getSelection ().setMarkerRange (new_range);
  }

  /**
   *  This method will ask the user for a range of bases (such as 100..200),
   *  using a TextRequester component, then selects that range.
   **/
  private void selectRange () {
    final TextRequester text_requester =
      new TextRequester ("enter a range of bases (eg. 100..200):", 18, "");

    text_requester.addTextRequesterListener (new TextRequesterListener () {
      public void actionPerformed (final TextRequesterEvent event) {
        final String range_string = event.getRequesterText ().trim ();

        if (range_string.length () == 0) {
          return;
        }

        final LocationLexer lexer = new LocationLexer (range_string);

        final TokenEnumeration enum = lexer.getTokens ();

        boolean complement_flag = false;

        if (enum.peekElement () instanceof String &&
            ((String)enum.peekElement ()).equals ("complement")) {
          complement_flag = true;
          enum.nextElement ();
        }

        enum.eatToken ('(');

        if (enum.peekElement () instanceof Integer) {
          int first_base = ((Integer) enum.nextElement ()).intValue ();

          if ((enum.eatToken ("..") || enum.eatToken ('.')) &&
              enum.peekElement () instanceof Integer) {
            int last_base = ((Integer) enum.nextElement ()).intValue ();

            enum.eatToken (')');

            if (enum.peekElement () == null) {
              try {
                final MarkerRange marker_range;
                if (complement_flag) {
                  final int temp = first_base;
                  first_base = last_base;
                  last_base = temp;
                }

                if (first_base <= last_base) {
                  // forward strand
                  final Strand strand;

                  strand = entry_group.getBases ().getForwardStrand ();

                  marker_range =
                    strand.makeMarkerRangeFromPositions (first_base,
                                                         last_base);
                } else {
                  // reverse strand
                  final Strand strand =
                    entry_group.getBases ().getBackwardStrand ();
                  final int raw_first =
                    entry_group.getBases ().getComplementPosition (first_base);
                  final int raw_last =
                    entry_group.getBases ().getComplementPosition (last_base);

                  marker_range =
                    strand.makeMarkerRangeFromPositions (raw_last,
                                                         raw_first);
                }

                getSelection ().setMarkerRange (marker_range);

                makeBaseVisible (getSelection ().getFirstBaseOfSelection ());

                return;
              } catch (OutOfRangeException e) {
                new MessageDialog (getParentFrame (),
                                   "the bases are out of range for this " +
                                   "sequence");
                return;
              }
            } else {
              new MessageDialog (getParentFrame (),
                                 "garbage at the end of the range: " + enum);
              return;
            }
          }
        }

        // if we get to here then there was a parse error
        new MessageDialog (getParentFrame (),
                           "error in range: the range should have this " +
                           "form: 100..200 - error at: " + enum);

      }
    });

    text_requester.setVisible (true);
  }

  /**
   *  This method sends an GotoEvent to all the GotoEvent listeners that will
   *  make the given base visible.
   **/
  private void makeBaseVisible (final Marker base_marker) {
    goto_event_source.gotoBase (base_marker);
  }

  /**
   *  The EntryGroup object that was passed to the constructor.
   **/
  private EntryGroup entry_group;

  /**
   *  The GotoEventSource object that was passed to the constructor.
   **/
  final GotoEventSource goto_event_source;

  private MenuItem all_item = null;
  private MenuItem none_item = null;
  private MenuItem selection_toggle_item = null;
  private MenuItem select_same_type_item = null;
  private MenuItem select_range_item = null;
  private MenuItem select_orf_item = null;
  private MenuItem select_by_key_item = null;
  private MenuItem select_cds_item = null;
}
